// (C) 1997, 1998 by Paolo Gatti
//
// $Id: Array2d.h,v 1.7 1998/11/01 21:45:28 paolo Exp $
// $Log: Array2d.h,v $
// Revision 1.7  1998/11/01 21:45:28  paolo
// R4 compatibility
//
// Revision 1.6  1998/01/06 11:41:23  paolo
// Fixed a bug in AddRow/Column reallocation
//
// Revision 1.5  1998/01/05 21:02:55  paolo
// First public release
//
// Revision 1.4  1997/12/22 10:16:42  paolo
// AddColumn/Row don't erase the new column
//
// Revision 1.3	1997/12/18 20:35:34	paolo
// Adapted from gcc to mwcc - mwcc's quirks prevent default arguments from
// working well. Solved by expliciting default args.
//
// Revision 1.2	1997/12/18 20:14:51	paolo
// Fully rewritten, added (Add|Rem)(Column|Row)[s], added Flatten() and
// constructor from flattened form
//

#ifndef __ARRAY2D_H
#define __ARRAY2D_H

#ifndef __STDIO_H
#include <stdio.h>
#endif /* __STDIO_H */

template <class Type> class Array2d
{
public:
	Array2d(int sizex, int sizey, bool force_size, Type *flattened);
	Array2d(int sizex, int sizey, bool force_size);
	Array2d(int sizex, int sizey);
	~Array2d();
	int SizeX();
	int SizeY();
	Type Get(int x, int y);

// The following member functions return non-zero in case of error
	int Set(int x, int y, Type what);
	Type Default();
	void SetDefault(Type what);
	void Clear();
	int Resize(int sizex, int sizey, bool force_size=false);
	int Resize(int sizex, int sizey);
	int AddColumns(int first, int last);
	int RemColumns(int first, int last);
	int AddRows(int first, int last);
	int RemRows(int first, int last);
	int AddColumn(int where);
	int RemColumn(int where);
	int AddRow(int where);
	int RemRow(int where);

// Copy the 2d array into a compacted 1d array
	Type *Flatten();

private:
	void init(int sizex, int sizey, bool force_size, Type *flattened);
	int size_x, size_y;
	int real_size_x, real_size_y;
	Type *items;
	Type def;
	int coords(int x, int y, int rsx) { return x + rsx * y; };
};

//-------------------------------------------------------------------------
// Member functions

//-------------------------------------------------------------------------
// Constructors/Destructor

template <class Type> inline void
Array2d<Type>::init
(int sizex, int sizey, bool force_size, Type *flattened)
{
	if (sizex < 1)
		size_x = 8;
	else
		size_x = sizex;

	if (size_x < 8 && !force_size)
		real_size_x = 8;
	else
		real_size_x = sizex;

	if (sizey < 1)
		size_y = 8;
	else
		size_y = sizey;

	if (size_y < 8 && !force_size)
		real_size_y = 8;
	else
		real_size_y = sizey;

	items = new Type[real_size_x*real_size_y];

	if (flattened)
		for (int y=0; y<size_y; y++)
			for (int x=0; x<size_x; x++)
	items[coords(x,y,real_size_x)] = flattened[coords(x,y,size_x)];
};

template <class Type> inline
Array2d<Type>::Array2d(int sizex, int sizey, bool force_size, Type *flat)
{
	init(sizex, sizey, force_size, flat);
}

template <class Type> inline
Array2d<Type>::Array2d(int sizex, int sizey, bool force_size)
{
	init(sizex, sizey, force_size, NULL);
}

template <class Type> inline
Array2d<Type>::Array2d(int sizex, int sizey)
{
	init(sizex, sizey, false, NULL);
}

template <class Type> inline Array2d<Type>::~Array2d()
{
	if (items)
		delete [] items;
}

//-------------------------------------------------------------------------
// Read array size

template <class Type> inline int Array2d<Type>::SizeX()
{
	return size_x;
}

template <class Type> inline int Array2d<Type>::SizeY()
{
	return size_y;
}

//-------------------------------------------------------------------------
// Read and write

template <class Type> inline Type Array2d<Type>::Get(int x, int y)
{
	if (x < 0 || x >= size_x || y < 0 || y >= size_y || !items)
		return def;
	else
		return items[coords(x,y,real_size_x)];
}

template <class Type> inline int Array2d<Type>::Set(int x, int y, Type what)
{
	if (x < 0 || x >= size_x || y < 0 || y >= size_y || !items)
		return 2;
	else {
		items[coords(x,y,real_size_x)] = what;
		return 0;
	}
}

//-------------------------------------------------------------------------
// Default item (used for out-of-bounds and for unset items)

template <class Type> inline Type Array2d<Type>::Default()
{
	return def;
}

template <class Type> inline void Array2d<Type>::SetDefault(Type what)
{
	def = what;
}

//-------------------------------------------------------------------------
// Fill the array with the default item

template <class Type> inline void Array2d<Type>::Clear()
{
	if (items)
		for (int x = 0; x < real_size_x * real_size_y; x++)
			items[x] = def;
}

//-------------------------------------------------------------------------
// Resize the array

template <class Type> inline int
Array2d<Type>::Resize(int sizex, int sizey, bool force_size)
{
	if (sizex < 1 || sizey < 1)
		return 2;

	if (force_size || sizex > real_size_x || sizey > real_size_y) {
		Type *newitems;
		int xm = (sizex < size_x) ? sizex : size_x;
		int ym = (sizey < size_y) ? sizey : size_y;
		newitems = new Type[sizex*sizey];
		if (!newitems)
			return 1;
		for (int x = 0; x < xm; x++)
			for (int y = 0; y < ym; y++)
				newitems[coords(x,y,sizex)] = items[coords(x,y,real_size_x)];
		delete [] items;
		items = newitems;
		real_size_x = sizex;
		real_size_y = sizey;
	}

	size_x = sizex;
	size_y = sizey;

	return 0;
}

template <class Type> inline int
Array2d<Type>::Resize(int sizex, int sizey)
{
	return Resize(sizex, sizey, false);
}

//-------------------------------------------------------------------------
// Add or remove a range of columns or rows, fill empty spaces w/ default

template <class Type> inline int
Array2d<Type>::AddColumns(int first, int last)
{
	int x, y, offset;

	if (first > last) {
		x = first; first = last; last = x;
	}
	if (first < 0)
		return 2;
	else if (first > size_x)
		first = size_x;

	offset = last - first + 1;

	if (size_x + offset > real_size_x) { /* Full resize needed */
		int new_size_x = size_x + offset;
		int new_real_size_x = new_size_x;
		Type *newitems = new Type[new_real_size_x * real_size_y];
		if (!newitems)
			return 1;
		for (y = 0; y < real_size_y; y++) {
			for (x = 0; x < first; x++)
				newitems[coords(x,y,new_real_size_x)] =
					items[coords(x,y,real_size_x)];
			for (; x <= last; x++)
				newitems[coords(x,y,new_real_size_x)] = def;
			for (; x < new_size_x; x++)
				newitems[coords(x,y,new_real_size_x)] =
					items[coords(x - offset, y, real_size_x)];
		}
		delete [] items;
		items = newitems;
		real_size_x = new_real_size_x;
		size_x = new_size_x;

	} else { /* Just needs some reshuffling */
		for (y = 0; y < size_y; y++) {
			for (x = size_x - 1; x >= first; x--)
				items[coords(x+offset,y,real_size_x)] =
					items[coords(x,y,real_size_x)];
			for (x = first; x <= last; x++)
				items[coords(x,y,real_size_x)] = def;
		}
		size_x += offset;
	}

	return 0;
}

template <class Type> inline int Array2d<Type>::RemColumns(int first, int last)
{
	int x, y, offset;

	if (first > last) {
		x = first; first = last; last = x;
	}
	if (first < 0) first = 0;
	if (last >= size_x) last = size_x - 1;
	if (last < 0) return 2;

	offset = last - first + 1;

	for (y = 0; y < size_y; y++)
		for (x = first; x < size_x - offset; x++)
			items[coords(x, y, real_size_x)] =
				items[coords(x+offset,y,real_size_x)];

	size_x -= offset;

	return 0;
}

template <class Type> inline int
Array2d<Type>::AddRows(int first, int last)
{
	int x, y, offset;

	if (first > last) {
		x = first; first = last; last = x;
	}
	if (first < 0)
		return 2;
	else if (first > size_y)
		first = size_y;

	offset = last - first + 1;

	if (size_y + offset > real_size_y) { /* Full resize needed */
		int new_size_y = size_y + offset;
		int new_real_size_y = new_size_y;
		Type *newitems = new Type[real_size_x * new_real_size_y];
		if (!newitems)
			return 1;
		for (x = 0; x < real_size_x; x++) {
			for (y = 0; y < first; y++)
				newitems[coords(x,y,real_size_x)] =
					items[coords(x,y,real_size_x)];
			for (; y <= last; y++)
				newitems[coords(x,y,real_size_x)] = def;
			for (; y < new_size_y; y++)
				newitems[coords(x,y,real_size_x)] =
					items[coords(x, y - offset, real_size_x)];
		}
		delete [] items;
		items = newitems;
		real_size_y = new_real_size_y;
		size_y = new_size_y;

	} else { /* Just needs some reshuffling */
		for (x = 0; x < size_x; x++) {
			for (y = size_y - 1; y >= first; y--)
				items[coords(x,y+offset,real_size_x)] =
					items[coords(x,y,real_size_x)];
			for (y = first; y <= last; y++)
				items[coords(x,y,real_size_x)] = def;
		}
		size_y += offset;
	}

	return 0;
}

template <class Type> inline int Array2d<Type>::RemRows(int first, int last)
{
	int x, y, offset;

	if (first > last) {
		x = first; first = last; last = x;
	}
	if (first < 0) first = 0;
	if (last >= size_y) last = size_y - 1;
	if (last < 0) return 2;

	offset = last - first + 1;

	for (x = 0; x < size_x; x++)
		for (y = first; y < size_y - offset; y++)
			items[coords(x, y, real_size_x)] =
				items[coords(x,y+offset,real_size_x)];

	size_y -= offset;

	return 0;
}

//-------------------------------------------------------------------------
// Duplicate or kill a single column/row

template <class Type> inline int Array2d<Type>::AddColumn(int where)
{
	int x,y;

	if (where < 0 || where > size_x)
		return 2;

	if (size_x >= real_size_x) { /* Full resize needed */
		int new_size_x = size_x + 1;
		int new_real_size_x = size_x + 4;
		Type *newitems = new Type[new_real_size_x * real_size_y];
		if (!newitems)
			return 1;
		for (y = 0; y < real_size_y; y++) {
			for (x = 0; x < where; x++)
				newitems[coords(x,y,new_real_size_x)] =
					items[coords(x,y,real_size_x)];
			newitems[coords(where,y,new_real_size_x)] = (where < real_size_x)
				? items[coords(where,y,real_size_x)] : def;
			x++;
			for (; x < new_size_x; x++)
				newitems[coords(x,y,new_real_size_x)] =
					items[coords(x-1,y,real_size_x)];
		}
		delete [] items;
		items = newitems;
		real_size_x = new_real_size_x;
		size_x = new_size_x;

	} else { /* Just needs some reshuffling */
		for (y = 0; y < size_y; y++) {
			for (x = size_x - 1; x >= where; x--)
				items[coords(x+1,y,real_size_x)] =
					items[coords(x,y,real_size_x)];
//			items[coords(where,y,real_size_x)] = def;
		}
		size_x++;
	}

	return 0;
}

template <class Type> inline int Array2d<Type>::RemColumn(int where)
{
	int x,y;

	if (where < 0 || where >= size_x)
		return 2;

	for (y = 0; y < size_y; y++)
		for (x = where + 1; x < size_x; x++)
			items[coords(x - 1, y, real_size_x)] =
				items[coords(x,y,real_size_x)];

	size_x--;

	return 0;
}

template <class Type> inline int Array2d<Type>::AddRow(int where)
{
	int x,y;

	if (where < 0 || where > size_y)
		return 2;

	if (size_y >= real_size_y) { /* Full resize needed */
		int new_size_y = size_y + 1;
		int new_real_size_y = size_y + 4;
		Type *newitems = new Type[real_size_x * new_real_size_y];
		if (!newitems)
			return 1;
		for (x = 0; x < real_size_x; x++) {
			for (y = 0; y < where; y++)
				newitems[coords(x,y,real_size_x)] =
					items[coords(x,y,real_size_x)];
			newitems[coords(x,where,real_size_x)] = (where < real_size_y)
				? items[coords(x,where,real_size_x)] : def;
			y++;
			for (; y < new_size_y; y++)
				newitems[coords(x,y,real_size_x)] =
					items[coords(x,y-1,real_size_x)];
		}
		delete [] items;
		items = newitems;
		real_size_y = new_real_size_y;
		size_y = new_size_y;

	} else { /* Just needs some reshuffling */
		for (x = 0; x < size_x; x++) {
			for (y = size_y - 1; y >= where; y--)
				items[coords(x,y+1,real_size_x)] =
					items[coords(x,y,real_size_x)];
//			items[coords(x,where,real_size_x)] = def;
		}
		size_y++;
	}

	return 0;
}

template <class Type> inline int Array2d<Type>::RemRow(int where)
{
	int x,y;

	if (where < 0 || where >= size_y)
		return 2;

	for (x = 0; x < size_x; x++)
		for (y = where + 1; y < size_y; y++)
			items[coords(x, y - 1, real_size_x)] =
				items[coords(x,y,real_size_x)];

	size_y--;

	return 0;
}

template <class Type> inline Type *Array2d<Type>::Flatten()
{
	Type *flat = new Type[size_x*size_y];

	if (flat)
		for (int y=0; y<size_y; y++)
			for (int x=0; x<size_x; x++)
	flat[coords(x,y,size_x)] = items[coords(x,y,real_size_x)];

	return flat;
}

#endif /* __ARRAY2D_H */

