// (C) 1997, 1998 by Paolo Gatti
//
// $Id: GridView.cpp,v 1.7 1998/11/01 21:45:04 paolo Exp $
// $Log: GridView.cpp,v $
// Revision 1.7  1998/11/01 21:45:04  paolo
// R4 compatibility
//
// Revision 1.6  1998/01/05 21:00:40  paolo
// Implemented LockSep and NewChild, stricter typing, first public release
//
// Revision 1.5  1998/01/03 22:16:09  paolo
// Fixed a few small bugs
// Added archival/instantiation
//
// Revision 1.4  1997/12/25 12:44:31  paolo
// Implemented Add/RemoveChild, ChildAt, subview resizing, etc.
// First functioning version
//
// Revision 1.3  1997/12/22 21:13:06  paolo
// Debugged; removed the possibility of moving vertical and horizontal borders
// at the same time, it was a huge source of bugs.
//
// Revision 1.2  1997/12/22 10:17:37  paolo
// Almost completely rewritten
//
// Revision 1.1  1997/12/14 20:24:32  paolo
// Initial revision

#include "GridView.h"

//-------------------------------------------------------------------------
// Constructor, destructor, archival and instantiation
//-------------------------------------------------------------------------

//-------------------------------------------------------------------------
// GridView constructor
GridView::GridView(
	BRect frame,
	const char *name,
	uint32 resizingMode = B_FOLLOW_NONE,
	uint32 flags = NULL,
	float bsize = 2,
	float minsep = 0) :
BView(	frame, name, resizingMode,
		flags|B_WILL_DRAW|B_FRAME_EVENTS)
{
	bordersize = (bsize < 2) ? 2 : bsize;
	minseparation = (minsep < 10 * bordersize) ? 10 * bordersize : minsep;
	SetViewColor(192, 192, 192);
	separators[0].AddItem(new float(bordersize - 1));
	separators[0].AddItem(new float(Bounds().right - bordersize + 1));
	separators[1].AddItem(new float(bordersize - 1));
	separators[1].AddItem(new float(Bounds().bottom - bordersize + 1));
	views = new Array2d<int32>(1,1);
	views->SetDefault(-1);
	views->Clear();
	locked = seplocked = 0;
	subviewnum = 1;
	views->Set(0,0,0);
	children.AddItem(NULL);
	NewChild(0);
}

//-------------------------------------------------------------------------
// GridView Archive
status_t GridView::Archive(BMessage *archive, bool deep = true) const
{
	int32 nx, ny, i, j;

	BView::Archive(archive, deep);
	archive->AddString("class", class_name(this));
	archive->AddString("class", __GRIDVIEW__);

	archive->AddFloat("bordersize", bordersize);
	archive->AddFloat("minseparation", minseparation);
	archive->AddInt32("locked", locked);
	archive->AddInt32("seplocked", seplocked);
	archive->AddInt32("subviewnum", subviewnum);

	for (nx = separators[0].CountItems() - 1, i = 1; i < nx; i++)
		archive->AddFloat("sepx", *(float *)separators[0].ItemAt(i));
	for (ny = separators[1].CountItems() - 1, i = 1; i < ny; i++)
		archive->AddFloat("sepy", *(float *)separators[1].ItemAt(i));

	for (j = 0; j < ny; j++)
		for (i = 0; i < nx; i++)
			archive->AddInt32("views", views->Get(i,j));

	return B_OK;
}

//-------------------------------------------------------------------------
// GridView Archive constructor
GridView::GridView(BMessage *archive) : BView(archive)
{
	int32 nx, ny, i, j, k, current, m, n;
	float f;
	BPoint p;
	BView *v;
	int32 *childx, *childy;

// Load general data
	archive->FindFloat("bordersize", &bordersize);
	if (bordersize < 2) bordersize = 2;

	archive->FindFloat("minseparation", &minseparation);
	if (minseparation < 10*bordersize) minseparation = 10 * bordersize;

	archive->FindInt32("locked", &locked);
	archive->FindInt32("seplocked", &seplocked);
	archive->FindInt32("subviewnum", &subviewnum);

// Load separators
	SetViewColor(192, 192, 192);
	separators[0].AddItem(new float(bordersize - 1));
	for (nx = 0; archive->FindFloat("sepx", nx, &f) == B_OK; nx++)
		separators[0].AddItem(new float(f));
	separators[0].AddItem(new float(Bounds().right - bordersize + 1));
	nx++;

	separators[1].AddItem(new float(bordersize - 1));
	for (ny = 0; archive->FindFloat("sepy", ny, &f) == B_OK; ny++)
		separators[1].AddItem(new float(f));
	separators[1].AddItem(new float(Bounds().bottom - bordersize + 1));
	ny++;
	
// Load children geometry
	childx = new int32 [subviewnum];
	childy = new int32 [subviewnum];
	for (i = 0; i < subviewnum; i++)
		childx[i] = -1;

	views = new Array2d<int32>(nx,ny);
	views->SetDefault(-1);
	views->Clear();

	for (j = k = 0; j < ny; j++)
		for (i = 0; i < nx; i++, k++) {
			current = archive->FindInt32("views", k);
			views->Set(i, j, current);
			childx[current] = i;
			childy[current] = j;
		}

// Setup children list

	for (k = current = 0; k < subviewnum; k++)
		if (childx[k] != -1) { // There's a subview...
	// p is a point inside the view
			p = BPoint(
				(*(float *)separators[0].ItemAt(childx[k]) +
				 *(float *)separators[0].ItemAt(childx[k]+1)) / 2,
				(*(float *)separators[1].ItemAt(childy[k]) +
				 *(float *)separators[1].ItemAt(childy[k]+1)) /2);
	// find the child view, if any
			v = NULL;
			for (n = 0, m = CountChildren(); n < m; n++)
				if (BView::ChildAt(n)->Frame().Contains(p)) {
					v = BView::ChildAt(n);
					break;
				}
			children.AddItem(v);
			childx[k] = current; // compactify later
	// increase views pointer
			current++;
		}

	// Remove holes in the children list
	for (i = 0; i < nx; i++)
		for (j = 0; j < ny; j++)
			if ((k = views->Get(i,j)) != -1)
				views->Set(i, j, childx[k]);

	subviewnum = current;

	delete [] childx;
	delete [] childy;
}

//-------------------------------------------------------------------------
// GridView Instantiate
GridView *GridView::Instantiate(BMessage *archive) 
{
	if (validate_instantiation(archive, __GRIDVIEW__))
		return new GridView(archive); 
		return NULL; 
}

//-------------------------------------------------------------------------
// GridView Destructor
GridView::~GridView()
{
	int32 x, dir;

	for (dir = 0; dir < 2; dir++) {
		x = separators[dir].CountItems();
		while(x--)
			delete (float *)separators[dir].RemoveItem(x);
	}

	delete views;
}

//-------------------------------------------------------------------------
// Class implementation
//-------------------------------------------------------------------------

//-------------------------------------------------------------------------
// GridView Draw
void GridView::Draw(BRect updateRect)
{
	int32 x, y, nx, ny, xx, yy;
	BRect subrect, intersect;
	int32 current, other1, other2;

// Update borders

	nx = separators[0].CountItems() - 1;
	ny = separators[1].CountItems() - 1;
	
	for (x = 0; x < nx; x++) {
		for (y = 0; y < ny; y++) {
			current = views->Get(x,y);

// Seek left-top corner
			other1 = views->Get(x-1,y);
			other2 = views->Get(x,y-1);
			if (current != other1 && current != other2) {
// Determine right-bottom corner
				for (xx = x+1, other1 = views->Get(xx,y);
					 other1 == current;
					 xx++, other1 = views->Get(xx,y))
					 	;
				for (yy = y+1, other2 = views->Get(x,yy);
					 other2 == current;
					 yy++, other2 = views->Get(x,yy))
					 	;

// Draw the child's border

				subrect.left   = *(float *)separators[0].ItemAt(x);
				subrect.top    = *(float *)separators[1].ItemAt(y);
				subrect.right  = *(float *)separators[0].ItemAt(xx);
				subrect.bottom = *(float *)separators[1].ItemAt(yy);

				SetDrawingMode(B_OP_COPY);
				subrect.InsetBy(bordersize, bordersize);
				SetHighColor(128, 128, 128);
				StrokeRect(subrect);
				subrect.InsetBy(1,1);
				SetHighColor(225, 225, 225);
				StrokeRect(subrect);
				
// Draw an empty child's interior
				if (!children.ItemAt(current)) {
					subrect.InsetBy(1,1);
					if (subrect.Intersects(updateRect))
						DrawEmptyChild(subrect, subrect & updateRect);
				}
			}
		}
	}
}

//-------------------------------------------------------------------------
// GridView GetCoords
// float coords[2]: location of click point, in pixels
// int32 idx[2]: location of click point, in regions (output)
// float minsep: min distance from a separator
// idx[0] values mean:
//	0  -> on the left border
//	1  -> in the first region
//	2  -> on the border between regions 0 and 1
//	...
//	2*howmany - 2 -> on the right border
// IDX[1] does the same vertically.
// Even values of IDX correspond to edges, odd values to subviews.
void GridView::GetCoords(float coords[2], int32 idx[2], float minsep) const
{
	int32 dir, n, howmany[2];
	float f;

	for (dir = 0; dir < 2; dir++) {
		howmany[dir] = separators[dir].CountItems();
		idx[dir] = -3;
		for(n=0; n<howmany[dir] && idx[dir] == -3; n++) {
			f=coords[dir] - *(float *)separators[dir].ItemAt(n);
			if (fabs(f) < minsep)
				idx[dir] = 2*n;
			else if(f < 0)
				idx[dir] = 2*n - 1;
		}
	}
}

//-------------------------------------------------------------------------
// GridView AlignCoords
// We might be aligned to a separator, but not on it...
void GridView::AlignCoords(int32 idx[2]) const
{
	int32 id1, id2;

	id1 = views->Get(idx[0] / 2, idx[1] / 2);
	if (id1 != -1) {
		if (!(idx[0]&1) && (idx[1]&1)) { // Vertical border
			id2 = views->Get(idx[0] / 2 - 1, idx[1] / 2);
			if (id1 == id2)
				idx[0]--;
		}
		else if ((idx[0]&1) && !(idx[1]&1)) { // Horizontal border
			id2 = views->Get(idx[0] / 2, idx[1] / 2 - 1);
			if (id1 == id2)
				idx[1]--;
		}
		else if (!(idx[0]&1)) { // Corner
			id2 = views->Get(idx[0] / 2 - 1, idx[1] / 2 - 1);
			if (id1 == id2)
				idx[0]--, idx[1]--;
		}
	}
}

//-------------------------------------------------------------------------
// GridView CalcSgmt
// Find selected edges (sgmt) from click coordinates (idx), return dragging
// direction (0 : drag vertical separator horizontally, 1 vice versa)
int32 GridView::CalcSgmt(int32 idx[2], int32 sgmt[2]) const
{
	int32 mode, n, id1, id2, tmp[2], howmany[2];

	for (n = 0; n < 2; n++)
		howmany[n] = separators[n].CountItems();

	if (idx[0]&1) {		// Horizontal border
		mode = 1;		// Move vertically
		sgmt[0] = idx[0] / 2;
		sgmt[1] = sgmt[0] + 1;
	}
	else if (idx[1]&1) {	// Vertical border
		mode = 0;			// Move horizontally
		sgmt[0] = idx[1] / 2;
		sgmt[1] = sgmt[0] + 1;
	}
	else if (!idx[0] && (!idx[1] || idx[1] == 2*howmany[1] - 2)) {
		mode = 1;	// Left corner - move vertically by default
		sgmt[0] = 0;
		sgmt[1] = 1;
	}
	else if (idx[0] == 2*howmany[0] - 2 &&
			 (!idx[1] || idx[1] == 2*howmany[1] - 2)) {
		mode = 1;	// Right corner - move vertically by default
		sgmt[1] = idx[0] / 2;
		sgmt[0] = sgmt[1] - 1;
	}
	else {	// internal corner: determine if it's a T or a +
		tmp[0] = idx[0] / 2;
		tmp[1] = idx[1] / 2;
		id1 = views->Get(tmp[0] - 1, tmp[1] - 1);
		id2 = views->Get(tmp[0] - 1, tmp[1]);
		if (id1 == id2) { // |- 
			mode = 0;
			sgmt[0] = tmp[1] - 1;
			sgmt[1] = sgmt[0] + 2;
		}
		else {
			id1 = views->Get(tmp[0], tmp[1] - 1);
			id2 = views->Get(tmp[0], tmp[1]);
			if (id1 == id2) { // -|
				mode = 0;
				sgmt[0] = tmp[1] - 1;
				sgmt[1] = sgmt[0] + 2;
			}
			else {	// + or horizontal T, move vertically by default
				mode = 1;
				sgmt[0] = tmp[0] - 1;
				sgmt[1] = sgmt[0] + 2;
			}
		}
	}

//	printf ("Moving separator %d (direction %d)\n", idx[mode] / 2, mode);

//	printf ("Min segment between separators %d-%d (of direction %d)\n",
//			sgmt[0], sgmt[1], 1 - mode);
			
	return mode;
}

//-------------------------------------------------------------------------
// GridView CalcToMov
// Extend moved edges
void GridView::CalcToMov(int32 mode, int32 sgmt[2], int32 idx[2], int32 mintomov[2],
						int32 maxtomov[2], uint32 buttons) const
{
	int32 id1, id2, tmp[2];

	mintomov[0] = mintomov[1] = sgmt[0];
	maxtomov[0] = maxtomov[1] = sgmt[1];

// Left button: move as little as possible
	if (buttons & B_PRIMARY_MOUSE_BUTTON) {

		tmp[1 - mode] = sgmt[0];
		tmp[mode] = idx[mode] / 2 - 1;
		id1 = views->Get(tmp[0], tmp[1]);
		if (id1 != -1) do {
			tmp[1 - mode]--;
			id2 = views->Get(tmp[0], tmp[1]);
			if (id1 == id2)
				mintomov[0]--;
			else
				break;
		} while (1);

		tmp[1 - mode] = sgmt[0];
		tmp[mode] = idx[mode] / 2;
		id1 = views->Get(tmp[0], tmp[1]);
		if (id1 != -1) do {
			tmp[1 - mode]--;
			id2 = views->Get(tmp[0], tmp[1]);
			if (id1 == id2)
				mintomov[1]--;
			else
				break;
		} while (1);

		tmp[1 - mode] = sgmt[1] - 1;
		tmp[mode] = idx[mode] / 2 - 1;
		id1 = views->Get(tmp[0], tmp[1]);
		if (id1 != -1) do {
			tmp[1 - mode]++;
			id2 = views->Get(tmp[0], tmp[1]);
			if (id1 == id2)
				maxtomov[0]++;
			else
				break;
		} while (1);

		tmp[1 - mode] = sgmt[1] - 1;
		tmp[mode] = idx[mode] / 2;
		id1 = views->Get(tmp[0], tmp[1]);
		if (id1 != -1) do {
			tmp[1 - mode]++;
			id2 = views->Get(tmp[0], tmp[1]);
			if (id1 == id2)
				maxtomov[1]++;
			else
				break;
		} while (1);
	}
	
// Right button: move as much as possible
	else if (buttons & B_SECONDARY_MOUSE_BUTTON) {
		do {
			tmp[1 - mode] = mintomov[0] - 1;
			tmp[mode] = idx[mode] / 2;
			id1 = views->Get(tmp[0], tmp[1]);
			tmp[mode]--;
			id2 = views->Get(tmp[0], tmp[1]);
			if (id1 != id2) {
				mintomov[0]--;
				mintomov[1]--;
			}
			else
				break;
		} while (1);

		do {
			tmp[1 - mode] = maxtomov[0];
			tmp[mode] = idx[mode] / 2;
			id1 = views->Get(tmp[0], tmp[1]);
			tmp[mode]--;
			id2 = views->Get(tmp[0], tmp[1]);
			if (id1 != id2) {
				maxtomov[0]++;
				maxtomov[1]++;
			}
			else
				break;
		} while (1);
	}
}

//-------------------------------------------------------------------------
// GridView CalcBounds
// Starting from drag bounds (minsgmt, maxsgmt) determine how far the
// separator can be dragged (xcoords)
void GridView::CalcBounds(int32 mode, int32 idx[2], int32 mintomov[2],
							int32 maxtomov[2], float xcoords[2]) const
{
	int32 bound[2], tmp[2], n, id1, id2;

	bound[0] = 0;
	bound[1] = separators[mode].CountItems() - 1;

	for (n = mintomov[0]; n < maxtomov[0]; n++) {
		tmp[1 - mode] = n;
		tmp[mode] = bound[0];
		id1 = views->Get(tmp[0], tmp[1]);
		while (++tmp[mode] < idx[mode] / 2) {
			id2 = views->Get(tmp[0], tmp[1]);
			if (id1 != id2) {
				id1 = id2;
				bound[0] = tmp[mode];
			}
		}
	}

	for (n = mintomov[1]; n < maxtomov[1]; n++) {
		tmp[1 - mode] = n;
		tmp[mode] = bound[1];
		id1 = views->Get(tmp[0], tmp[1]);
		while (--tmp[mode] >= idx[mode] / 2) {
			id2 = views->Get(tmp[0], tmp[1]);
			if (id1 != id2) {
				id1 = id2;
				bound[1] = tmp[mode] + 1;
			}
		}
	}

	xcoords[0] = *(float *)separators[mode].ItemAt(bound[0]);
	xcoords[1] = *(float *)separators[mode].ItemAt(bound[1]);

//	printf ("Moveable between separators %d-%d (dir %d, pix %f %f)\n",
//			bound[0], bound[1], mode, xcoords[0], xcoords[1]);
}

//-------------------------------------------------------------------------
// GridView CommitChanges
// Perform changes to the grid
void GridView::CommitChanges(int32 mode, int32 idx[2], int32 newidx[2],
								int32 mintomov[2], int32 maxtomov[2],
								float coords[2])
{
	int32 n, tmp[2], howmany[2], id1, id2, add, mntm[2], mxtm[2];
	BView *view;
	float delta[2];
	
	for (n = 0; n < 2; n++)
		howmany[n] = separators[n].CountItems();

// From margin to margin or on self: nothing happens
	if (idx[mode] == newidx[mode] ||
		((idx[mode] == 0 || idx[mode] == 2*howmany[mode] - 2) &&
		 (newidx[mode] == 0 || newidx[mode] == 2*howmany[mode] - 2))) {
//		printf("No change.\n");
		return;
	}

// If separators are locked, extend the motion to the next separator

	if (seplocked && (newidx[mode]&1))
	{
		if (newidx[mode] < idx[mode])
			newidx[mode]--;
		else
			newidx[mode]++;
	}

// Add a new separator or move it, stage one: duplicate the row/column...
// (Problem w/ moving: can't move a short distance...)

	else if (newidx[mode]&1) {
		n = newidx[mode] / 2;
		if (mode)
			views->AddRow(n);
		else
			views->AddColumn(n);
		separators[mode].AddItem(new float(coords[mode]), n+1);
		for (tmp[1-mode]=0; tmp[1-mode]<howmany[1-mode]-1; tmp[1-mode]++) {
			tmp[mode] = n;
			id1 = views->Get(tmp[0], tmp[1]);
			tmp[mode]++;
			views->Set(tmp[0], tmp[1], id1);
		}
		howmany[mode]++;
		newidx[mode]++;
		if (idx[mode] >= newidx[mode])
			idx[mode] += 2;

//		printf("Added separator: DIR %d IDX %d NEWIDX %d\n",
//				mode, idx[mode], newidx[mode]);
	}

// Determine whether to add a new view (add = true) or not
	if (idx[mode] == 0 || idx[mode] == 2*howmany[mode] - 2)
		add = true;
	else if (newidx[mode] < idx[mode]) {
		if ((mintomov[0] > mintomov[1] || maxtomov[0] < maxtomov[1]))
			add = true;
		else {
			tmp[0] = mintomov[0];
			tmp[1] = maxtomov[0];
			CalcToMov(mode, tmp, idx, mntm, mxtm, B_PRIMARY_MOUSE_BUTTON);
			add = (mntm[0] > mntm[1] || mxtm[0] < mxtm[1]);
		}
	}
	else {
		if ((mintomov[0] > mintomov[1] || maxtomov[0] < maxtomov[1]))
			add = true;
		else {
			tmp[0] = mintomov[1];
			tmp[1] = maxtomov[1];
			CalcToMov(mode, tmp, idx, mntm, mxtm, B_PRIMARY_MOUSE_BUTTON);
			add = (mntm[0] < mntm[1] || mxtm[0] > mxtm[1]);
		}
	}

// Determine deltas
	delta[mode] = fabs(*(float *)separators[mode].ItemAt(newidx[mode] / 2) -
						*(float *)separators[mode].ItemAt(idx[mode] / 2));
	delta[1 - mode] = 0;

// Move the view border
	if (add) {
		add = subviewnum++;
		children.AddItem(NULL);
	}
//	printf ("ADDVIEW %d\n", add);

	id1 = id2 = -1;	// Last resized view, decreased/increased

	if (newidx[mode] > idx[mode]) {
		for (tmp[1 - mode] = mintomov[1];
			 tmp[1 - mode] < maxtomov[1];
			 tmp[1 - mode]++) {
			// Decrease size of existing subviews to the right
			tmp[mode] = idx[mode] / 2;
			n = views->Get(tmp[0], tmp[1]);
			if (id1 != n) {
				id1 = n;
				view = (BView *)children.ItemAt(n);
				if (view) {
					if (delta[0] < view->Frame().Width() &&
						delta[1] < view->Frame().Height())
					{
						view->ResizeBy(-delta[0], -delta[1]);
						view->MoveBy(delta[0], delta[1]);
						view->Invalidate();
					}
					else
						RemoveChild(n);
				}
			}
			if (add) {
				// Change subview codes to new view
				for (;	tmp[mode] < newidx[mode] / 2;
						tmp[mode]++)
					views->Set(tmp[0], tmp[1], add);
			}
			else
			{
				tmp[mode] = idx[mode] / 2 - 1;
				n = views->Get(tmp[0], tmp[1]);
				for (tmp[mode] = idx[mode] / 2;
						tmp[mode] < newidx[mode] / 2;
						tmp[mode]++) {
					views->Set(tmp[0], tmp[1], n);
				}
				// Increase size of existing subviews to the right
				tmp[mode] = idx[mode] / 2;
				n = views->Get(tmp[0], tmp[1]);
				if (id2 != n) {
					id2 = n;
					view = (BView *)children.ItemAt(n);
					if (view) {
						view->ResizeBy(delta[0], delta[1]);
						view->Invalidate();
					}
				}
			}
		}
	}
	else if (newidx[mode] < idx[mode]) {
		for (tmp[1 - mode] = mintomov[0];
			 tmp[1 - mode] < maxtomov[0];
			 tmp[1 - mode]++) {
			// Decrease size of existing subviews to the left
			tmp[mode] = idx[mode] / 2 - 1;
			n = views->Get(tmp[0], tmp[1]);
			if (id1 != n) {
				id1 = n;
				view = (BView *)children.ItemAt(id1);
				if (view) {
					if (delta[0] < view->Frame().Width() &&
						delta[1] < view->Frame().Height()) {
							view->ResizeBy(-delta[0], -delta[1]);
							view->Invalidate();
					}
					else
						RemoveChild(n);
				}
			}
			if (add) {
				for (tmp[mode] = idx[mode] / 2 - 1;
						tmp[mode] >= newidx[mode] / 2;
						tmp[mode]--)
					views->Set(tmp[0], tmp[1], add);
			}
			else
			{
				tmp[mode] = idx[mode] / 2;
				n = views->Get(tmp[0], tmp[1]);
				for (tmp[mode] = idx[mode] / 2 - 1;
						tmp[mode] >= newidx[mode] / 2;
						tmp[mode]--) {
					views->Set(tmp[0], tmp[1], n);
				}
				// Increase size of existing subviews to the right
				tmp[mode] = idx[mode] / 2;
				n = views->Get(tmp[0], tmp[1]);
				if (id2 != n) {
					id2 = n;
					view = (BView *)children.ItemAt(n);
					if (view) {
						view->MoveBy(-delta[0], -delta[1]);
						view->ResizeBy(delta[0], delta[1]);
						view->Invalidate();
					}
				}
			}
		}
	}

// Remove the old separator and its orthogonals if they're empty
	if (!seplocked) {
		RemoveEmptySeparator(mode, idx[mode] / 2, howmany);

		for (n = howmany[1 - mode] - 2; n > 0; n--)
			RemoveEmptySeparator(1 - mode, n, howmany);
	}

// Notify creation of a new child
	if (add)
		NewChild(add);

	Invalidate();
}

//-------------------------------------------------------------------------
// GridView RemoveEmptySeparator
void GridView::RemoveEmptySeparator(int32 mode, int32 which, int32 howmany[2])
{
	int32 tmp[2], id1, id2;

	for (tmp[1-mode]=0; tmp[1-mode]<howmany[1-mode]-1; tmp[1-mode]++) {
		tmp[mode] = which;
		id1 = views->Get(tmp[0], tmp[1]);
		tmp[mode]--;
		id2 = views->Get(tmp[0], tmp[1]);
		if (id1 != id2)
			return;
	}
	
	// Separator empty: remove
	if (mode)
		views->RemRow(which);
	else
		views->RemColumn(which);
	float *fp = (float *)separators[mode].RemoveItem(which);
	delete fp;
	howmany[mode]--;
}

//-------------------------------------------------------------------------
// GridView MouseDown
inline void GridView::DrawCursorX(float where)
{
	StrokeLine(
		BPoint(where - 2, 1),
		BPoint(where + 2, 1));
	StrokeLine(
		BPoint(where - 2, Bounds().bottom - 1),
		BPoint(where + 2, Bounds().bottom - 1));
}

inline void GridView::DrawCursorY(float where)
{
	StrokeLine(
		BPoint(1, where - 2),
		BPoint(1, where + 2));
	StrokeLine(
		BPoint(Bounds().right - 1, where - 2),
		BPoint(Bounds().right - 1, where + 2));
}

void GridView::MouseDown(BPoint where)
{
	BPoint locat;
	uint32 buttons;
	int32 idx[2], newidx[2], dir, howmany[2], mode, id1;
	int32 mintomov[2], maxtomov[2], sgmt[2];
	float coords[2], xcoords[2], origcoords[2],
			minmvcrdl, maxmvcrdl, minmvcrdr, maxmvcrdr;
	
	GetMouse(&locat, &buttons, false);
//	printf("MOUSEDOWN %lu %f %f\n", buttons, where.x, where.y);
	if (!buttons) return;

	for (dir = 0; dir < 2; dir++)
		howmany[dir] = separators[dir].CountItems();

	SetDrawingMode(B_OP_INVERT);

// First of all, determine where we are
	coords[0] = where.x;
	coords[1] = where.y;
	GetCoords(coords, idx, bordersize + 2);
	AlignCoords(idx);
//	printf("IDX: %d %d\n", idx[0], idx[1]);

// Are we inside a subview?

	if ((idx[0]&1) && (idx[1]&1)) {
		id1 = views->Get(idx[0] / 2, idx[1] / 2);
		if (id1 != -1 && !children.ItemAt(id1))
			ClickOnChild(where, idx[0] / 2, idx[1] / 2);
		return;
	}

// User clicked on a border: begin resizing
	if (locked) return;				// Can't drag locked stuff

// What to move?
	mode = CalcSgmt(idx, sgmt);
	CalcToMov(mode, sgmt, idx, mintomov, maxtomov, buttons);

	minmvcrdl = *(float *)separators[1 - mode].ItemAt(mintomov[0]);
	maxmvcrdl = *(float *)separators[1 - mode].ItemAt(maxtomov[0]);
	minmvcrdr = *(float *)separators[1 - mode].ItemAt(mintomov[1]);
	maxmvcrdr = *(float *)separators[1 - mode].ItemAt(maxtomov[1]);
	origcoords[mode] = *(float *)separators[mode].ItemAt(idx[mode] / 2);
	origcoords[1-mode] = *(float *)separators[1-mode].ItemAt(idx[1-mode]/2);

//	printf ("Draw cursor between: MINL %d MAXL %d MINR %d MAXR %d\n",
//			mintomov[0], maxtomov[0], mintomov[1], maxtomov[1]);

// Determine motion bounds

	CalcBounds(mode, idx, mintomov, maxtomov, xcoords);

// Show what's going to move and in which directions (cursors)

	SetDrawingMode(B_OP_INVERT);

	if (!mode) {
		if (idx[0])
			StrokeLine(
				BPoint(origcoords[0] - 1, minmvcrdl + 2),
				BPoint(origcoords[0] - 1, maxmvcrdl - 2));
				
		if (idx[0] != 2*howmany[0] - 2)
			StrokeLine(
				BPoint(origcoords[0] + 1, minmvcrdr + 2),
				BPoint(origcoords[0] + 1, maxmvcrdr - 2));
				
		where.x = origcoords[0];
		DrawCursorX(where.x);
	}
	else {
		if (idx[1])
			StrokeLine(
				BPoint(minmvcrdl + 2, origcoords[1] - 1),
				BPoint(maxmvcrdl - 2, origcoords[1] - 1));
		if (idx[1] != 2*howmany[1] - 2)
			StrokeLine(
				BPoint(minmvcrdr + 2, origcoords[1] + 1),
				BPoint(maxmvcrdr - 2, origcoords[1] + 1));
				
		where.y = origcoords[1];
		DrawCursorY(where.y);
	}

// Now keep track of where the mouse is going...

	while(buttons) {
		snooze(20000);
		GetMouse(&locat, &buttons, false);

		if (!mode) {
			if (locat.x < xcoords[0])
				locat.x = xcoords[0];
			else if (locat.x > xcoords[1])
				locat.x = xcoords[1];
			if (locat.x != where.x) {
				DrawCursorX(where.x);
				where.x = locat.x;
				DrawCursorX(where.x);
			}
		}
		else {
			if (locat.y < xcoords[0])
				locat.y = xcoords[0];
			else if (locat.y > xcoords[1])
				locat.y = xcoords[1];
			if (locat.y != where.y) {
				DrawCursorY(where.y);
				where.y = locat.y;
				DrawCursorY(where.y);
			}
		}
	}

// Remove cursors
	
	if (!mode) {
		if (idx[0])
			StrokeLine(
				BPoint(origcoords[0] - 1, minmvcrdl + 2),
				BPoint(origcoords[0] - 1, maxmvcrdl - 2));
				
		if (idx[0] != 2*howmany[0] - 2)
			StrokeLine(
				BPoint(origcoords[0] + 1, minmvcrdr + 2),
				BPoint(origcoords[0] + 1, maxmvcrdr - 2));

		DrawCursorX(where.x);
	}
	else{
		if (idx[1])
			StrokeLine(
				BPoint(minmvcrdl + 2, origcoords[1] - 1),
				BPoint(maxmvcrdl - 2, origcoords[1] - 1));
		if (idx[1] != 2*howmany[1] - 2)
			StrokeLine(
				BPoint(minmvcrdr + 2, origcoords[1] + 1),
				BPoint(maxmvcrdr - 2, origcoords[1] + 1));

		DrawCursorY(where.y);
	}

// Where did the user release?

	coords[0] = where.x;
	coords[1] = where.y;
	GetCoords(coords, newidx, minseparation);
//	printf("NEWIDX %d %d COORD %f %f\n",
//			newidx[0], newidx[1], coords[0], coords[1]);

// Commit changes

	CommitChanges(mode, idx, newidx, mintomov, maxtomov, coords);
}

//-------------------------------------------------------------------------
// DragSeparator: emulate mouse dragging programmatically
void GridView::DragSeparator(int32 idx[2], float coords[2], uint32 buttons)
{
	BPoint locat;
	int32 newidx[2], dir, howmany[2], mode;
	int32 mintomov[2], maxtomov[2], sgmt[2];
	float xcoords[2];
	
	if (!buttons) return;

	for (dir = 0; dir < 2; dir++)
		howmany[dir] = separators[dir].CountItems();

// First of all, determine where we are
//	printf("IDX: %d %d\n", idx[0], idx[1]);

// Are we inside a subview?

	if ((idx[0]&1) && (idx[1]&1))
		return;

// User clicked on a border: begin resizing
	if (locked) return;				// Can't drag locked stuff

// What to move?
	mode = CalcSgmt(idx, sgmt);
	CalcToMov(mode, sgmt, idx, mintomov, maxtomov, buttons);

//	printf ("Draw cursor between: MINL %d MAXL %d MINR %d MAXR %d\n",
//			mintomov[0], maxtomov[0], mintomov[1], maxtomov[1]);

// Determine motion bounds

	CalcBounds(mode, idx, mintomov, maxtomov, xcoords);

// Now keep track of where the mouse is going...

		if (!mode) {
			if (coords[0] < xcoords[0])
				coords[0] = xcoords[0];
			else if (coords[0] > xcoords[1])
				coords[0] = xcoords[1];
		}
		else {
			if (coords[1] < xcoords[0])
				coords[1] = xcoords[0];
			else if (coords[1] > xcoords[1])
				coords[1] = xcoords[1];
	}

// Where did the user release?

	GetCoords(coords, newidx, minseparation);
//	printf("NEWIDX %d %d COORD %f %f\n",
//			newidx[0], newidx[1], coords[0], coords[1]);

// Commit changes

	CommitChanges(mode, idx, newidx, mintomov, maxtomov, coords);
}

//-------------------------------------------------------------------------
// AddChild - Fill a child view
void GridView::AddChild(BView *v, int32 id)
{
	int32 x, xx, y, yy, nx, ny, id2;

	RemoveChild(id);		// Make sure the place is empty
	
// Find subview coordinates
	nx = separators[0].CountItems() - 1;
	ny = separators[1].CountItems() - 1;

// Seek left-top corner
	for (x = 0; x < nx; x++) {
		for (y = 0; y < ny; y++) {
			id2 = views->Get(x,y);
			if (id == id2)
				break;
		}
		if (id == id2)
			break;
	}

// Determine right-bottom corner
	for (xx = x+1, id2 = views->Get(xx,y);
		 id == id2;
		 xx++, id2 = views->Get(xx,y))
		 	;
	for (yy = y+1, id2 = views->Get(x,yy);
		 id == id2;
		 yy++, id2 = views->Get(x,yy))
		 	;

// Resize child and add it
	v->ResizeTo(	*(float *)separators[0].ItemAt(xx)
					- *(float *)separators[0].ItemAt(x) - 2*bordersize - 4,
					*(float *)separators[1].ItemAt(yy)
					- *(float *)separators[1].ItemAt(y) - 2*bordersize - 4);
	v->MoveTo(	*(float *)separators[0].ItemAt(x) + bordersize + 2,
				*(float *)separators[1].ItemAt(y) + bordersize + 2);
	children.RemoveItem(id);
	children.AddItem(v, id);
	BView::AddChild(v);
}

//-------------------------------------------------------------------------
// Virtual functions
//-------------------------------------------------------------------------

//-------------------------------------------------------------------------
// NewChild - what to do if the user creates a subview?
void GridView::NewChild(int32 id)
{
	;
}

//-------------------------------------------------------------------------
// DrawEmptyChild - What to draw? frame -> entire child, updateRect -> part
void GridView::DrawEmptyChild(BRect frame, BRect updateRect)
{
// Fill with uniform grey
	SetDrawingMode(B_OP_COPY);
	SetHighColor(192,192,192);
	FillRect(updateRect);
}

//-------------------------------------------------------------------------
// ClickOnChild - User clicked inside an empty child view
void GridView::ClickOnChild(BPoint where, int32 x, int32 y)
{
	;
}

//-------------------------------------------------------------------------
// ChildDetached - what to do if the user detaches a subview?
void GridView::ChildDetached(BView *view)
{
	;
}

